from __future__ import annotations

import re
import os
from base64 import b64encode
from pathlib import Path as P

import click

from .component import component_parameters_key
from .config import Config
from .helpers import rm_tree_contents, yaml_dump


class SecretRef:
    """
    Helper class for finding Kapitan secret ref strings and producing Kapitan
    secret ref files
    """

    _SECRET_REF = re.compile(r"\?{([^}]+)\}")

    def __init__(self, key, ref):
        self.keys = [key]
        refelems = ref.split(":")
        self.type = refelems[0]
        self.ref = refelems[1]

    def __str__(self):
        return f"SecretRef(invkeys={self.keys}, type={self.type}, ref={self.ref})"

    @property
    def refstr(self):
        return f"{self.type}:{self.ref}"

    @classmethod
    def from_value(cls, key, value):
        """
        Create SecretRef object from string `value`, if the string contains a
        Kapitan secret reference. If no secret reference is contained in
        `value` this method returns None.
        """
        m = cls._SECRET_REF.search(value)
        if m:
            return SecretRef(key, m.group(1))

        return None

    def _mangle_ref(self):
        """
        Transform Kapitan secret reference into base64 encoded reference for
        kapitan refs --reveal. This method replaces one part of what
        `kapitan refs --write` would do.

        Currently only `vaultkv` references are supported.
        """
        if self.type == "vaultkv":
            secret, key = self.ref.rsplit("/", 1)
            return b64encode(f"{secret}:{key}".encode()).decode("utf-8")

        raise NotImplementedError(f"Ref type: {self.type}")

    def create_kapitan_ref(self, refdir, ref_params, debug=False):
        """
        Create a Kapitan secret ref file from this reference.

        This method creates a secret ref file in `refdir` which Kapitan can
        subsequently use for revealing the secret.

        Currently only `vaultkv` references are supported.
        """
        reffile = P(refdir, self.ref)
        if debug:
            click.echo(f"    > Writing to file {reffile}")
        if self.type != "vaultkv":
            raise NotImplementedError(f"ref type: {self.type}")
        params = ref_params[self.type]
        refdef = {
            "data": self._mangle_ref(),
            "encoding": "original",
            "type": self.type,
            params["key"]: params["values"],
        }
        os.makedirs(reffile.parent, exist_ok=True)
        yaml_dump(refdef, reffile)

    def add_key(self, key):
        self.keys.append(key)


class RefBuilder:
    """
    Helper class to wrap recursive search for Kapitan secret references
    """

    _refs: dict[str, SecretRef]

    def __init__(self, config: Config, inventory):
        self.debug = config.debug
        self.trace = config.trace
        self._bootstrap_target = config.inventory.bootstrap_target
        self.inventory = inventory
        self._refs = {}
        self._ref_params = None

    def _find_ref(self, key, value):
        """
        Process leaf value of parameters structure.
        """
        # Only consider leaves which are of type string, other types cannot
        # contain a secret reference.
        if isinstance(value, str):
            r = SecretRef.from_value(key, value)
            if r is not None:
                if self.debug:
                    click.echo(f"    > Found secret ref {r.refstr} in {value}")
                if r.refstr in self._refs:
                    if self.trace:
                        click.echo("    > Duplicate ref, adding key to list")
                    self._refs[r.refstr].add_key(key)
                else:
                    self._refs[r.refstr] = r
        elif self.trace:
            click.echo(f"    > Ignoring leaf of type {type(value).__name__}...")

    def _find_refs(self, prefix, params):
        """
        Recursively search Kapitan refs, descending into dicts and lists.
        """
        if self.trace:
            click.echo(f" > Processing {prefix}")

        if isinstance(params, dict):
            # Recurse for dicts
            for k, v in params.items():
                self._find_refs(f"{prefix}/{k}", v)
        elif isinstance(params, list):
            # Recurse for lists
            for idx, e in enumerate(params):
                self._find_refs(f"{prefix}[{idx}]", e)
        else:
            # Otherwise, handle as leaf
            self._find_ref(prefix, params)

    def find_refs(self, target: str, key: str):
        """
        Search for Kapitan secret refs in key `key` in the parameters for
        Kapitan target `target`.
        """
        params = self.inventory[target]["parameters"][key]
        self._find_refs(key, params)

    @property
    def refs(self):
        return self._refs.values()

    @property
    def params(self):
        if self._ref_params is None:
            kapitan_params = self.inventory[self._bootstrap_target]["parameters"][
                "kapitan"
            ]["secrets"]["vaultkv"]
            self._ref_params = {
                "vaultkv": {"key": "vault_params", "values": kapitan_params}
            }
        return self._ref_params


def update_refs(config: Config, aliases: dict[str, str], inventory: dict):
    """
    Iterate over parameters for each target, and create Kapitan secret refs
    for all ?{...} found as values in the dicts
    """
    click.secho("Updating Kapitan secret references...", bold=True)
    os.makedirs(config.refs_dir, exist_ok=True)
    rm_tree_contents(config.refs_dir)

    rb = RefBuilder(config, inventory)

    # Generate list of component keys from component aliases dict
    component_keys = set(map(component_parameters_key, aliases.values()))
    bootstrap_target = config.inventory.bootstrap_target
    # Find all keys in the bootstrap target's parameters which don't directly
    # belong to a component.
    non_component_keys = (
        set(inventory[bootstrap_target]["parameters"].keys()) - component_keys
    )
    # Search those keys in the bootstrap target for secret references
    for key in non_component_keys:
        rb.find_refs(bootstrap_target, key)

    for target, component in aliases.items():
        # For components, any dashes in the component name are replaced by
        # underscores in the parameters key.
        component_key = component_parameters_key(component)
        # Find references for component instance
        rb.find_refs(target, component_key)

    ref_params = rb.params
    # Create Kapitan references
    for r in rb.refs:
        if config.debug:
            click.echo(f" > Creating Kapitan reffile for secret ref {r.refstr}")
        r.create_kapitan_ref(config.refs_dir, ref_params, debug=config.debug)



from __future__ import annotations

import click

from commodore.config import Config
from commodore.component import Component, component_dir
from commodore.helpers import relsymlink
from commodore.package import Package, package_dependency_dir

from .component_library import validate_component_library_name
from .discovery import _discover_components, _discover_packages
from .tools import format_component_list
from .version_parsing import _read_components, _read_packages


def create_component_symlinks(cfg, component: Component):
    """
    Create symlinks in the inventory subdirectory.

    The actual code for components lives in the dependencies/ subdirectory, but
    we want to access some of their files through the inventory.
    """
    relsymlink(component.class_file, cfg.inventory.components_dir)
    inventory_default = cfg.inventory.defaults_file(component)
    relsymlink(
        component.defaults_file,
        inventory_default.parent,
        dest_name=inventory_default.name,
    )

    for file in component.lib_files:
        if cfg.debug:
            click.echo(f"     > installing template library: {file}")
        relsymlink(
            validate_component_library_name(component.name, file),
            cfg.inventory.lib_dir,
        )


def create_package_symlink(cfg, pname: str, package: Package):
    """
    Create package symlink in the inventory.

    Packages are downloaded to `dependencies/pkg.{package-name}/{path}` and symlinked to
    `inventory/classes/{package-name}`.
    """
    if not package.target_dir:
        raise ValueError("Can't symlink package which doesn't have a working directory")
    if not package.target_dir.is_dir():
        raise ValueError(
            f"Can't symlink package subpath {package.sub_path}, not a directory"
        )
    relsymlink(package.target_dir, cfg.inventory.classes_dir, dest_name=pname)


def fetch_components(cfg: Config):

    """
    Download all components required by target.

    This function discovers required components by parsing key `applications` in the
    hierarchy.
    """

    click.secho("Discovering components...", bold=True)
    cfg.inventory.ensure_dirs()
    component_names, component_aliases = _discover_components(cfg)
    click.secho("Registering component aliases...", bold=True)
    cfg.register_component_aliases(component_aliases)
    cspecs = _read_components(cfg, component_names)
    click.secho("Fetching components...", bold=True)
    for cn in component_names:
        cspec = cspecs[cn]
        if cfg.debug:
            click.echo(f" > Fetching component {cn}...")
        cdep = cfg.register_dependency_repo(cspec.url)
        c = Component(
            cn,
            work_dir=cfg.work_dir,
            dependency=cdep,
            version=cspec.version,
            sub_path=cspec.path,
        )
        c.checkout()
        cfg.register_component(c)
        create_component_symlinks(cfg, c)


def register_components(cfg: Config):
    """
    Discover components in the inventory, and register them if the
    corresponding directory in `dependencies/` exists.

    Create component symlinks for discovered components which exist.
    """
    click.secho("Discovering included components...", bold=True)
    try:
        components, component_aliases = _discover_components(cfg)
        cspecs = _read_components(cfg, components)
    except KeyError as e:
        raise click.ClickException(f"While discovering components: {e}")
    click.secho("Registering components and aliases...", bold=True)

    for cn in components:
        cspec = cspecs[cn]
        if cfg.debug:
            click.echo(f" > Registering component {cn}...")
        if not component_dir(cfg.work_dir, cn).is_dir():
            click.secho(
                f" > Skipping registration of component {cn}: repo is not available",
                fg="yellow",
            )
            continue
        cdep = cfg.register_dependency_repo(cspec.url)
        component = Component(
            cn,
            work_dir=cfg.work_dir,
            dependency=cdep,
            sub_path=cspec.path,
            version=cspec.version,
        )
        cfg.register_component(component)
        create_component_symlinks(cfg, component)

    registered_components = cfg.get_components().keys()
    pruned_aliases = {
        a: c for a, c in component_aliases.items() if c in registered_components
    }
    pruned = sorted(set(component_aliases.keys()) - set(pruned_aliases.keys()))
    if len(pruned) > 0:
        click.secho(
            f" > Dropping alias(es) {pruned} with missing component(s).", fg="yellow"
        )
    cfg.register_component_aliases(pruned_aliases)


def fetch_packages(cfg: Config):
    """
    Download configuration packages used by the cluster.

    This function discovers packages which are used by parsing key `applications` in the
    hierarchy.
    """

    click.secho("Discovering config packages...", bold=True)
    cfg.inventory.ensure_dirs()
    pkgs = _discover_packages(cfg)
    pspecs = _read_packages(cfg, pkgs)

    for p in pkgs:
        pspec = pspecs[p]
        pdep = cfg.register_dependency_repo(pspec.url)
        pkg = Package(
            p,
            dependency=pdep,
            target_dir=package_dependency_dir(cfg.work_dir, p),
            version=pspec.version,
            sub_path=pspec.path,
        )
        pkg.checkout()
        cfg.register_package(p, pkg)
        create_package_symlink(cfg, p, pkg)


def register_packages(cfg: Config):
    """
    Discover configuration packages used by the cluster in the inventory and register
    them if they're checked out in `inventory/classes`.

    This function discovers packages which are used by parsing key `applications` in the
    hierarchy.
    """

    click.secho("Discovering config packages...", bold=True)
    cfg.inventory.ensure_dirs()
    pkgs = _discover_packages(cfg)
    pspecs = _read_packages(cfg, pkgs)
    for p in pkgs:
        pkg_dir = package_dependency_dir(cfg.work_dir, p)
        if not pkg_dir.is_dir():
            click.secho(
                f" > Skipping registration of package '{p}': repo is not available",
                fg="yellow",
            )
            continue
        pspec = pspecs[p]
        pdep = cfg.register_dependency_repo(pspec.url)
        pkg = Package(p, dependency=pdep, target_dir=pkg_dir, sub_path=pspec.path)
        cfg.register_package(p, pkg)
        create_package_symlink(cfg, p, pkg)


def verify_version_overrides(cluster_parameters):
    errors = []
    for cname, cspec in cluster_parameters["components"].items():
        if "url" not in cspec:
            errors.append(f"component '{cname}'")

    for pname, pspec in cluster_parameters.get("packages", {}).items():
        if "url" not in pspec:
            errors.append(f"package '{pname}'")

    if len(errors) > 0:
        names = format_component_list(errors, format_func=lambda c: c)

        s = "s" if len(errors) > 1 else ""
        have = "have" if len(errors) > 1 else "has"

        raise click.ClickException(
            f"Version override{s} specified for {names} which {have} no URL"
        )



from __future__ import annotations

import json

from pathlib import Path as P

import _jsonnet
import click

from commodore import __install_dir__
from commodore.config import Config
from commodore.component import Component

from .jsonnet import jsonnet_runner


def _output_dir(work_dir: P, instance: str, path):
    """Compute directory in which to apply filter"""
    return work_dir / "compiled" / instance / path


def _builtin_filter_helm_namespace(
    work_dir: P, inv, component: Component, instance: str, path, **kwargs
):
    if "namespace" not in kwargs:
        raise click.ClickException(
            "Builtin filter 'helm_namespace': filter argument 'namespace' is required"
        )
    create_namespace = kwargs.get("create_namespace", "false")
    # Transform create_namespace to string as jsonnet extvars can only be
    # strings
    if isinstance(create_namespace, bool):
        create_namespace = "true" if create_namespace else "false"
    exclude_objects = kwargs.get("exclude_objects", [])
    exclude_objects = "|".join([json.dumps(e) for e in exclude_objects])
    output_dir = _output_dir(work_dir, instance, path)

    # pylint: disable=c-extension-no-member
    jsonnet_runner(
        work_dir,
        inv,
        component.name,
        instance,
        path,
        _jsonnet.evaluate_file,
        __install_dir__ / "filters" / "helm_namespace.jsonnet",
        namespace=kwargs["namespace"],
        create_namespace=create_namespace,
        exclude_objects=exclude_objects,
        chart_output_dir=str(output_dir),
    )


_builtin_filters = {
    "helm_namespace": _builtin_filter_helm_namespace,
}


class UnknownBuiltinFilter(ValueError):
    def __init__(self, filtername):
        super().__init__(f"Unknown builtin filter: {filtername}")
        self.filtername = filtername


# pylint: disable=too-many-arguments
def run_builtin_filter(
    config: Config,
    inv: dict,
    component: Component,
    instance: str,
    filterid: str,
    path: P,
    **filterargs: str,
):
    if filterid not in _builtin_filters:
        raise UnknownBuiltinFilter(filterid)
    _builtin_filters[filterid](
        config.work_dir, inv, component, instance, path, **filterargs
    )


# pylint: disable=unused-argument
def validate_builtin_filter(config: Config, c: Component, instance: str, fd: dict):
    if fd["filter"] not in _builtin_filters:
        raise UnknownBuiltinFilter(fd["filter"])

    if "filterargs" not in fd:
        raise KeyError("Builtin filter is missing required key 'filterargs'")

    fpath = _output_dir(config.work_dir, instance, fd["path"])
    if not fpath.exists():
        raise ValueError("Builtin filter called on path which doesn't exist")



from __future__ import annotations

from collections.abc import Iterable
from dataclasses import dataclass

from enum import Enum

import click

from commodore.config import Config
from commodore.helpers import kapitan_inventory


class DepType(Enum):
    COMPONENT = "components"
    PACKAGE = "packages"


class DependencyParseError(ValueError):
    field: str

    def __init__(self, field: str):
        super().__init__("Error parsing dependency specification")
        self.field = field


@dataclass
class DependencySpec:
    """Class for parsed Dependency specification"""

    url: str
    version: str
    path: str

    @classmethod
    def parse(cls, info: dict[str, str]) -> DependencySpec:
        if "url" not in info:
            raise DependencyParseError("url")

        if "version" not in info:
            raise DependencyParseError("version")

        path = info.get("path", "")
        if path.startswith("/"):
            path = path[1:]

        return DependencySpec(info["url"], info["version"], path)


def _read_versions(
    cfg: Config,
    dependency_type: DepType,
    dependency_names: Iterable[str],
    require_key: bool = True,
    ignore_class_notfound: bool = False,
) -> dict[str, DependencySpec]:
    deps_key = dependency_type.value
    deptype_str = dependency_type.name.lower()
    deptype_cap = deptype_str.capitalize()
    dependencies = {}

    inv = kapitan_inventory(cfg, ignore_class_notfound=ignore_class_notfound)
    cluster_inventory = inv[cfg.inventory.bootstrap_target]
    deps = cluster_inventory["parameters"].get(deps_key, None)
    if not deps:
        if require_key:
            raise click.ClickException(
                f"{deptype_cap} list ('parameters.{deps_key}') missing"
            )
        # If we don't require the key for the requested dependency type to be present,
        # just set deps to the empty dict.
        deps = {}

    for depname in dependency_names:
        if depname not in deps:
            raise click.ClickException(
                f"Unknown {deptype_str} '{depname}'."
                + f" Please add it to 'parameters.{deps_key}'"
            )

        try:
            dep = DependencySpec.parse(deps[depname])
        except DependencyParseError as e:
            raise click.ClickException(
                f"{deptype_cap} '{depname}' is missing field '{e.field}'"
            )

        if cfg.debug:
            click.echo(f" > URL for {depname}: {dep.url}")
            click.echo(f" > Version for {depname}: {dep.version}")
            click.echo(f" > Subpath for {depname}: {dep.path}")

        dependencies[depname] = dep

    return dependencies


def _read_components(
    cfg: Config, component_names: Iterable[str]
) -> dict[str, DependencySpec]:
    return _read_versions(cfg, DepType.COMPONENT, component_names)


def _read_packages(
    cfg: Config, package_names: Iterable[str]
) -> dict[str, DependencySpec]:
    return _read_versions(
        cfg,
        DepType.PACKAGE,
        package_names,
        require_key=False,
        ignore_class_notfound=True,
    )



from __future__ import annotations

from pathlib import Path
from shutil import rmtree
from typing import Sequence

import click
import git

from cookiecutter.main import cookiecutter

from commodore import __install_dir__
from commodore.component import Component, component_dir
from commodore.dependency_templater import Templater, Renderer
from commodore.multi_dependency import MultiDependency


class ComponentTemplater(Templater):
    library: bool
    post_process: bool
    matrix_tests: bool

    @property
    def cookiecutter_args(self) -> dict[str, str]:
        return {
            "add_lib": "y" if self.library else "n",
            "add_pp": "y" if self.post_process else "n",
            "add_golden": "y" if self.golden_tests else "n",
            "add_matrix": "y" if self.matrix_tests else "n",
            "copyright_holder": self.copyright_holder,
            "copyright_year": self.today.strftime("%Y"),
            "github_owner": self.github_owner,
            "name": self.name,
            "slug": self.slug,
            "release_date": self.today.strftime("%Y-%m-%d"),
        }

    @property
    def target_dir(self) -> Path:
        return component_dir(self.config.work_dir, self.slug)

    @property
    def template(self) -> str:
        component_template = __install_dir__ / "component-template"
        return str(component_template.resolve())

    @property
    def deptype(self) -> str:
        return "component"

    @property
    def template_renderer(self) -> Renderer:
        return cookiecutter

    @property
    def repo_url(self) -> str:
        return f"git@github.com:{self.github_owner}/component-{self.slug}.git"

    @property
    def additional_files(self) -> Sequence[str]:
        return [
            ".github",
            ".gitignore",
            ".*.yml",
            ".editorconfig",
        ]

    def delete(self):
        cdir = component_dir(self.config.work_dir, self.slug)
        if cdir.exists():
            cr = git.Repo(cdir)
            cdep = MultiDependency(
                cr.remote().url, self.config.inventory.dependencies_dir
            )
            component = Component(
                self.slug, dependency=cdep, work_dir=self.config.work_dir
            )

            if not self.config.force:
                click.confirm(
                    "Are you sure you want to delete component "
                    f"{self.slug}? This action cannot be undone",
                    abort=True,
                )
            rmtree(component.target_directory)
            # We check for other checkouts here, because our MultiDependency doesn't
            # know if there's other dependencies which would be registered on it.
            if not cdep.has_checkouts():
                # Also delete bare copy of component repo, if there's no other
                # worktree checkouts for the same dependency repo.
                rmtree(cdep.repo_directory)
            else:
                click.echo(
                    f" > Not deleting bare copy of repository {cdep.url}. "
                    + "Other worktrees refer to the same reposiotry."
                )

            click.secho(f"Component {self.slug} successfully deleted 🎉", bold=True)
        else:
            raise click.BadParameter(
                "Cannot find component with slug " f"'{self.slug}'."
            )



from __future__ import annotations

import collections
import itertools
import json
import shutil
import os
from collections.abc import Callable, Iterable
from pathlib import Path as P
from typing import Optional

import click
import requests
import yaml

# pylint: disable=redefined-builtin
from requests.exceptions import ConnectionError, HTTPError
from url_normalize import url_normalize
from kapitan import cached
from kapitan import targets as kapitan_targets
from kapitan import defaults
from kapitan.cached import reset_cache as reset_reclass_cache
from kapitan.refs.base import RefController, PlainRef
from kapitan.refs.secrets.vaultkv import VaultBackend
from kapitan.resources import inventory_reclass

from commodore import __install_dir__
from commodore.config import Config


ArgumentCache = collections.namedtuple("ArgumentCache", ["inventory_path"])


class FakeVaultBackend(VaultBackend):
    def __init__(self):
        "init FakeVaultBackend ref backend type"
        super().__init__(None)

    def __getitem__(self, ref_path):
        return PlainRef(ref_path)


class ApiError(Exception):
    pass


def yaml_load(file):
    """
    Load single-document YAML and return document
    """
    with open(file, "r", encoding="utf-8") as f:
        return yaml.safe_load(f)


def yaml_load_all(file):
    """
    Load multi-document YAML and return documents in list
    """
    with open(file, "r", encoding="utf-8") as f:
        return list(yaml.safe_load_all(f))


def _represent_str(dumper, data):
    """
    Custom string rendering when dumping data as YAML.

    Hooking this method into PyYAML with

        yaml.add_representer(str, _represent_str)

    will configure the YAML dumper to render strings which contain newline
    characters as block scalars with the last newline stripped.
    """
    style = None
    if "\n" in data:
        style = "|"
    return dumper.represent_scalar("tag:yaml.org,2002:str", data, style=style)


def yaml_dump(obj, file):
    """
    Dump obj as single-document YAML
    """
    yaml.add_representer(str, _represent_str)
    with open(file, "w", encoding="utf-8") as outf:
        yaml.dump(obj, outf)


def yaml_dump_all(obj, file):
    """
    Dump obj as multi-document YAML
    """
    yaml.add_representer(str, _represent_str)
    with open(file, "w", encoding="utf-8") as outf:
        yaml.dump_all(obj, outf)


def lieutenant_query(api_url, api_token, api_endpoint, api_id):
    try:
        r = requests.get(
            url_normalize(f"{api_url}/{api_endpoint}/{api_id}"),
            headers={"Authorization": f"Bearer {api_token}"},
        )
    except ConnectionError as e:
        raise ApiError(f"Unable to connect to Lieutenant at {api_url}") from e
    try:
        resp = json.loads(r.text)
    except json.JSONDecodeError as e:
        raise ApiError("Client error: Unable to parse JSON") from e
    try:
        r.raise_for_status()
    except HTTPError as e:
        extra_msg = "."
        if r.status_code >= 400:
            if "reason" in resp:
                extra_msg = f": {resp['reason']}"
            else:
                extra_msg = f": {e}"
        raise ApiError(f"API returned {r.status_code}{extra_msg}") from e
    else:
        return resp


def _verbose_rmtree(tree, *args, **kwargs):
    click.echo(f" > deleting {tree}/")
    shutil.rmtree(tree, *args, **kwargs)


def clean_working_tree(config: Config):
    # Defining rmtree as a naked Callable means that mypy won't complain about
    # _verbose_rmtree and shutil.rmtree having slightly different signatures.
    rmtree: Callable
    if config.debug:
        rmtree = _verbose_rmtree
    else:
        rmtree = shutil.rmtree
    click.secho("Cleaning working tree", bold=True)
    rmtree(config.inventory.inventory_dir, ignore_errors=True)
    rmtree(config.inventory.lib_dir, ignore_errors=True)
    rmtree(config.inventory.libs_dir, ignore_errors=True)
    rmtree(config.inventory.output_dir, ignore_errors=True)
    rmtree(config.catalog_dir, ignore_errors=True)


# pylint: disable=too-many-arguments
def kapitan_compile(
    config: Config,
    targets: Iterable[str],
    output_dir: P = None,
    search_paths=None,
    fake_refs=False,
    reveal=False,
):
    if not output_dir:
        output_dir = config.work_dir

    if not search_paths:
        search_paths = []
    search_paths = search_paths + [
        config.work_dir,
        __install_dir__,
    ]
    reset_reclass_cache()
    refController = RefController(config.refs_dir)
    if fake_refs:
        refController.register_backend(FakeVaultBackend())
    click.secho("Compiling catalog...", bold=True)
    cached.args["compile"] = ArgumentCache(
        inventory_path=config.inventory.inventory_dir
    )
    kapitan_targets.compile_targets(
        inventory_path=config.inventory.inventory_dir,
        search_paths=search_paths,
        output_path=output_dir,
        targets=targets,
        parallel=4,
        labels=None,
        ref_controller=refController,
        verbose=config.trace,
        prune=False,
        indent=2,
        reveal=reveal,
        cache=False,
        cache_paths=None,
        fetch_dependencies=config.fetch_dependencies,
        force_fetch=True,
        validate=False,
        schemas_path=config.work_dir / "schemas",
        jinja2_filters=defaults.DEFAULT_JINJA2_FILTERS_PATH,
    )


def kapitan_inventory(
    config: Config, key: str = "nodes", ignore_class_notfound: bool = False
) -> dict:
    """
    Reset reclass cache and render inventory.
    Returns the top-level key according to the kwarg.
    """
    reset_reclass_cache()
    inv = inventory_reclass(
        config.inventory.inventory_dir, ignore_class_notfound=ignore_class_notfound
    )
    return inv[key]


def rm_tree_contents(basedir):
    """
    Delete all files in directory `basedir`, but do not delete the directory
    itself.
    """
    basedir = P(basedir)
    if not basedir.is_dir():
        raise ValueError("Expected directory as argument")
    for f in basedir.glob("*"):
        if f.name.startswith("."):
            # pathlib's glob doesn't filter hidden files, skip them here
            continue
        if f.is_dir():
            shutil.rmtree(f)
        else:
            os.unlink(f)


# pylint: disable=unsubscriptable-object
def relsymlink(src: P, dest_dir: P, dest_name: Optional[str] = None):
    if dest_name is None:
        dest_name = src.name
    # pathlib's relative_to() isn't suitable for this use case, since it only
    # works for dropping a path's prefix according to the documentation. See
    # https://docs.python.org/3/library/pathlib.html#pathlib.PurePath.relative_to
    link_src = os.path.relpath(src, start=dest_dir)
    link_dst = dest_dir / dest_name
    if not P(src).exists():
        raise click.ClickException(
            f"Can't link {link_src} to {link_dst}. Source does not exist."
        )
    if link_dst.exists() or link_dst.is_symlink():
        os.remove(link_dst)
    os.symlink(link_src, link_dst)


def sliding_window(iterable, n):
    # sliding_window('ABCDEFG', 4) -> ABCD BCDE CDEF DEFG
    it = iter(iterable)
    window = collections.deque(itertools.islice(it, n), maxlen=n)
    if len(window) == n:
        yield tuple(window)
    for x in it:
        window.append(x)
        yield tuple(window)



from __future__ import annotations

import os

from typing import Any, Optional, Union

import click

from .helpers import (
    lieutenant_query,
    yaml_dump,
    yaml_load,
)

from .component import component_parameters_key, Component
from .config import Config
from .inventory import Inventory


class Cluster:
    _cluster_response: dict
    _tenant_response: dict
    _fallback_dynamic_facts: dict[str, Any]

    def __init__(
        self,
        cluster_response: dict,
        tenant_response: dict,
        fallback_dynamic_facts: dict[str, Any] = {},
    ):
        self._cluster = cluster_response
        self._tenant = tenant_response
        self._fallback_dynamic_facts = fallback_dynamic_facts

        if (
            "tenant" not in self._cluster
            or self._cluster["tenant"] != self._tenant["id"]
        ):
            raise click.ClickException("Tenant ID mismatch")

    @property
    def id(self) -> str:
        return self._cluster["id"]

    @property
    def display_name(self) -> str:
        return self._cluster["displayName"]

    @property
    def global_git_repo_url(self) -> str:
        field = "globalGitRepoURL"
        if field not in self._tenant:
            raise click.ClickException(
                f"URL of the global git repository is missing on tenant '{self.tenant_id}'"
            )
        return self._tenant[field]

    def _extract_field(self, field: str, default) -> str:
        """
        Extract `field` from the tenant and cluster data, preferring the value present in the
        cluster data over the value in the tenant data. If field is not present in both tenant and
        cluster data, return `default`.
        """
        return self._cluster.get(field, self._tenant.get(field, default))

    @property
    def global_git_repo_revision(self) -> str:
        return self._extract_field("globalGitRepoRevision", None)

    @property
    def config_repo_url(self) -> str:
        repo_url = self._tenant.get("gitRepo", {}).get("url", None)
        if repo_url is None:
            raise click.ClickException(
                f" > API did not return a repository URL for tenant '{self._cluster['tenant']}'"
            )
        return repo_url

    @property
    def config_git_repo_revision(self) -> str:
        return self._extract_field("tenantGitRepoRevision", None)

    @property
    def catalog_repo_url(self) -> str:
        repo_url = self._cluster.get("gitRepo", {}).get("url", None)
        if repo_url is None:
            raise click.ClickException(
                f" > API did not return a repository URL for cluster '{self._cluster['id']}'"
            )
        return repo_url

    @property
    def tenant_id(self) -> str:
        return self._tenant["id"]

    @property
    def tenant_display_name(self) -> str:
        return self._tenant["displayName"]

    @property
    def facts(self) -> dict[str, str]:
        return self._cluster.get("facts", {})

    @property
    def dynamic_facts(self) -> dict[str, Any]:
        if "dynamicFacts" in self._cluster and self._fallback_dynamic_facts:
            empty = "" if self._cluster["dynamicFacts"] else "empty "
            click.secho(
                f" > Cluster API response contains {empty}dynamic facts, ignoring "
                + " dynamic facts provided on the command line."
            )

        return self._cluster.get("dynamicFacts", self._fallback_dynamic_facts)


def load_cluster_from_api(cfg: Config, cluster_id: str) -> Cluster:
    cluster_response = lieutenant_query(
        cfg.api_url, cfg.api_token, "clusters", cluster_id
    )
    if "tenant" not in cluster_response:
        raise click.ClickException("cluster does not have a tenant reference")
    tenant_response = lieutenant_query(
        cfg.api_url, cfg.api_token, "tenants", cluster_response["tenant"]
    )
    return Cluster(cluster_response, tenant_response, cfg.dynamic_facts)


def read_cluster_and_tenant(inv: Inventory) -> tuple[str, str]:
    """
    Reads the cluster and tenant ID from the current target.
    """
    file = inv.params_file
    if not file.is_file():
        raise click.ClickException(f"params file for {file.stem} does not exist")

    data = yaml_load(file)

    return (
        data["parameters"][inv.bootstrap_target]["name"],
        data["parameters"][inv.bootstrap_target]["tenant"],
    )


def render_target(
    inv: Inventory,
    target: str,
    components: dict[str, Component],
    # pylint: disable=unsubscriptable-object
    component: Optional[str] = None,
):
    if not component:
        component = target
    bootstrap = target == inv.bootstrap_target
    if not bootstrap and component not in components:
        raise click.ClickException(f"Target {target} is not a component")

    classes = [f"params.{inv.bootstrap_target}"]
    parameters: dict[str, Union[dict, str]] = {
        "_instance": target,
    }
    if not bootstrap:
        parameters["_base_directory"] = str(components[component].target_directory)

    for c in components:
        if inv.defaults_file(c).is_file():
            classes.append(f"defaults.{c}")
        else:
            click.secho(f" > Default file for class {c} missing", fg="yellow")

    classes.append("global.commodore")

    if not bootstrap:
        if not inv.component_file(component).is_file():
            raise click.ClickException(
                f"Target rendering failed for {target}: component class is missing"
            )
        classes.append(f"components.{component}")
        parameters["kapitan"] = {
            "vars": {
                "target": target,
            },
        }

        # When component != target we're rendering a target for an aliased
        # component. This needs some extra work.
        if component != target:
            ckey = component_parameters_key(component)
            tkey = component_parameters_key(target)
            parameters[tkey] = {}
            parameters[ckey] = f"${{{tkey}}}"

    return {
        "classes": classes,
        "parameters": parameters,
    }


# pylint: disable=unsubscriptable-object
def update_target(cfg: Config, target: str, component: Optional[str] = None):
    click.secho(f"Updating Kapitan target for {target}...", bold=True)
    file = cfg.inventory.target_file(target)
    os.makedirs(file.parent, exist_ok=True)
    targetdata = render_target(
        cfg.inventory, target, cfg.get_components(), component=component
    )
    yaml_dump(targetdata, file)


def render_params(inv: Inventory, cluster: Cluster):
    facts = cluster.facts
    dynfacts = cluster.dynamic_facts
    for fact in ["distribution", "cloud"]:
        if fact not in facts or not facts[fact]:
            raise click.ClickException(f"Required fact '{fact}' not set")

    data = {
        "parameters": {
            inv.bootstrap_target: {
                "name": cluster.id,
                "display_name": cluster.display_name,
                "catalog_url": cluster.catalog_repo_url,
                "tenant": cluster.tenant_id,
                "tenant_display_name": cluster.tenant_display_name,
            },
            "facts": facts,
            "dynamic_facts": dynfacts,
        },
    }

    return data


def update_params(inv: Inventory, cluster: Cluster):
    click.secho("Updating cluster parameters...", bold=True)
    file = inv.params_file
    os.makedirs(file.parent, exist_ok=True)
    yaml_dump(render_params(inv, cluster), file)



from __future__ import annotations

import json
import os
import functools

from collections.abc import Callable, Iterable
from pathlib import Path as P
from typing import Any

import _jsonnet

from commodore.config import Config
from commodore.component import Component
from commodore.helpers import yaml_load, yaml_load_all, yaml_dump, yaml_dump_all
from commodore import __install_dir__


def _try_path(basedir: P, rel: str):
    """
    Returns content of file basedir/rel if it exists, None if file not found, or throws an exception
    """
    if not rel:
        raise RuntimeError("Got invalid filename (empty string).")
    if rel[0] == "/":
        full_path = P(rel)
    else:
        full_path = basedir / rel
    if full_path.is_dir():
        raise RuntimeError("Attempted to import a directory")

    if not full_path.is_file():
        return full_path.name, None
    with open(full_path, encoding="utf-8") as f:
        return full_path.name, f.read()


def _import_callback_with_searchpath(search: Iterable[P], basedir: P, rel: str):
    full_path, content = _try_path(basedir, rel)
    if content:
        return full_path, content
    for p in search:
        full_path, content = _try_path(p, rel)
        if content:
            return full_path, content
    raise RuntimeError("File not found")


def _import_cb(work_dir: P, basedir: str, rel: str):
    # Add current working dir to search path for Jsonnet import callback
    search_path = [
        work_dir.resolve(),
        __install_dir__.resolve(),
        (work_dir / "vendor").resolve(),
    ]
    return _import_callback_with_searchpath(search_path, P(basedir), rel)


def _list_dir(basedir: os.PathLike, basename: bool):
    """
    Non-recursively list files in directory `basedir`. If `basename` is set to
    True, only return the file name itself and not the full path.
    """
    files = [x for x in P(basedir).iterdir() if x.is_file()]

    if basename:
        return [f.parts[-1] for f in files]

    return files


_native_callbacks = {
    "yaml_load": (("file",), yaml_load),
    "yaml_load_all": (("file",), yaml_load_all),
    "list_dir": (
        (
            "dir",
            "basename",
        ),
        _list_dir,
    ),
}


def write_jsonnet_output(output_dir: P, output: str):
    out_objs = json.loads(output)
    for outobj, outcontents in out_objs.items():
        outpath = output_dir / f"{outobj}.yaml"
        if not outpath.exists():
            print(f"   > {outpath} doesn't exist, creating...")
            os.makedirs(outpath.parent, exist_ok=True)
        if isinstance(outcontents, list):
            yaml_dump_all(outcontents, outpath)
        else:
            yaml_dump(outcontents, outpath)


# pylint: disable=too-many-arguments
def jsonnet_runner(
    work_dir: P,
    inv: dict[str, Any],
    component: str,
    instance: str,
    path: os.PathLike,
    jsonnet_func: Callable,
    jsonnet_input: os.PathLike,
    **kwargs: str,
):
    def _inventory() -> dict[str, Any]:
        return inv

    _native_cb = _native_callbacks
    _native_cb["commodore_inventory"] = ((), _inventory)
    kwargs["target"] = component
    kwargs["component"] = component
    output_dir = work_dir / "compiled" / instance / path
    kwargs["output_path"] = str(output_dir)
    output = jsonnet_func(
        str(jsonnet_input),
        import_callback=functools.partial(_import_cb, work_dir),
        native_callbacks=_native_cb,
        ext_vars=kwargs,
    )
    write_jsonnet_output(output_dir, output)


def _filter_file(component: Component, filterpath: str) -> P:
    return component.target_directory / filterpath


def run_jsonnet_filter(
    config: Config,
    inv: dict,
    component: Component,
    instance: str,
    filterid: str,
    path: P,
    **filterargs: str,
):
    """
    Run user-supplied jsonnet as postprocessing filter. This is the original
    way of doing postprocessing filters.
    """
    filterfile = _filter_file(component, filterid)
    # pylint: disable=c-extension-no-member
    jsonnet_runner(
        config.work_dir,
        inv,
        component.name,
        instance,
        path,
        _jsonnet.evaluate_file,
        filterfile,
        **filterargs,
    )


# pylint: disable=unused-argument
def validate_jsonnet_filter(config: Config, c: Component, instance: str, fd: dict):
    filterfile = _filter_file(c, fd["filter"])
    if not filterfile.is_file():
        raise ValueError("Jsonnet filter definition does not exist")



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['commodore',
 'commodore.component',
 'commodore.component-template.hooks',
 'commodore.cruft',
 'commodore.cruft._commands',
 'commodore.cruft._commands.utils',
 'commodore.dependency_mgmt',
 'commodore.inventory',
 'commodore.package',
 'commodore.postprocess']

package_data = \
{'': ['*'],
 'commodore': ['component-template/*',
               'component-template/{{ cookiecutter.slug }}/*',
               'component-template/{{ cookiecutter.slug }}/.github/*',
               'component-template/{{ cookiecutter.slug '
               '}}/.github/ISSUE_TEMPLATE/*',
               'component-template/{{ cookiecutter.slug }}/.github/workflows/*',
               'component-template/{{ cookiecutter.slug }}/class/*',
               'component-template/{{ cookiecutter.slug }}/component/*',
               'component-template/{{ cookiecutter.slug }}/docs/*',
               'component-template/{{ cookiecutter.slug '
               '}}/docs/modules/ROOT/pages/*',
               'component-template/{{ cookiecutter.slug '
               '}}/docs/modules/ROOT/pages/explanations/*',
               'component-template/{{ cookiecutter.slug '
               '}}/docs/modules/ROOT/pages/how-tos/*',
               'component-template/{{ cookiecutter.slug '
               '}}/docs/modules/ROOT/pages/references/*',
               'component-template/{{ cookiecutter.slug '
               '}}/docs/modules/ROOT/pages/tutorials/*',
               'component-template/{{ cookiecutter.slug '
               '}}/docs/modules/ROOT/partials/*',
               'component-template/{{ cookiecutter.slug }}/lib/*',
               'component-template/{{ cookiecutter.slug }}/tests/*',
               'component-template/{{ cookiecutter.slug '
               '}}/tests/golden/defaults/{{ cookiecutter.slug }}/apps/*',
               'filters/*',
               'lib/*']}

install_requires = \
['PyGithub==1.55',
 'click==8.1.3',
 'cookiecutter==2.1.1',
 'gitpython==3.1.24',
 'importlib-metadata==4.12.0',
 'kapitan==0.30.0',
 'oauthlib==3.1.1',
 'pyjwt==2.1.0',
 'python-dotenv==0.20.0',
 'pyxdg==0.28',
 'requests[use_chardet_on_py3]==2.26.0',
 'typer==0.6.1',
 'url-normalize==1.4.3']

entry_points = \
{'console_scripts': ['commodore = commodore.cli:main',
                     'compile = tools.tools:compile',
                     'local_reveal = tools.tools:reveal']}

setup_kwargs = {
    'name': 'syn-commodore',
    'version': '1.7.0',
    'description': 'Commodore provides opinionated tenant-aware management of Kapitan inventories and templates. Commodore uses Kapitan for the heavy lifting of rendering templates and resolving a hierachical configuration structure.',
    'long_description': '# Project Syn: Commodore\n\n[![Docker](https://github.com/projectsyn/commodore/actions/workflows/push.yml/badge.svg)](https://github.com/projectsyn/commodore/actions/workflows/push.yml)\n[![PyPI release](https://github.com/projectsyn/commodore/actions/workflows/publish-pypi.yml/badge.svg)](https://github.com/projectsyn/commodore/actions/workflows/publish-pypi.yml)\n[![GitHub Release](https://img.shields.io/github/v/release/projectsyn/commodore.svg)](https://github.com/projectsyn/commodore/releases)\n[![PyPI Release](https://img.shields.io/pypi/v/syn-commodore?color=blue)](https://pypi.org/project/syn-commodore)\n[![Maintainability](https://api.codeclimate.com/v1/badges/abb63d489a6d6e01939d/maintainability)](https://codeclimate.com/github/projectsyn/commodore/maintainability)\n[![Test Coverage](https://api.codeclimate.com/v1/badges/abb63d489a6d6e01939d/test_coverage)](https://codeclimate.com/github/projectsyn/commodore/test_coverage)\n\nThis repository is part of Project Syn.\nFor documentation on Project Syn and this component, see https://syn.tools.\n\n\nSee [GitHub Releases](https://github.com/projectsyn/commodore/releases) for changelogs of each release version of Commodore.\n\nSee [DockerHub](https://hub.docker.com/r/projectsyn/commodore) for pre-built Docker images of Commodore.\n\nCommodore is [published on PyPI](https://pypi.org/project/syn-commodore/)\n\n## Overview\n\nCommodore provides opinionated tenant-aware management of [Kapitan](https://kapitan.dev/) inventories and templates.\nCommodore uses Kapitan for the heavy lifting of rendering templates and resolving a hierachical configuration structure.\n\nCommodore introduces the concept of a component, which is a bundle of Kapitan templates and associated Kapitan classes which describe how to render the templates.\nCommodore fetches any components that are required for a given configuration before running Kapitan, and sets up symlinks so Kapitan can find the component classes.\n\nCommodore also supports additional processing on the output of Kapitan, such as patching in the desired namespace for a Helm chart which has been rendered using `helm template`.\n\n## System Requirements\n\n* Python 3.8 - 3.10 with `python3-dev` and `python3-venv` updated\n* [jsonnet-bundler](https://github.com/jsonnet-bundler/jsonnet-bundler)\n\n## Getting started\n\n1. Recommended: create a new virtual environment\n    ```console\n    python3 -m venv venv\n    source venv/bin/activate\n    ```\n1. Install commodore from PyPI\n    ```console\n    pip install syn-commodore\n    ```\n1. <a name="getting_started_jsonnet"></a>Install jsonnet-bundler according to upstream [documentation](https://github.com/jsonnet-bundler/jsonnet-bundler#install).\n\n1. For Commodore to work, you need to run an instance of [Lieutenant](https://syn.tools/syn/tutorials/getting-started.html#_kickstart_lieutenant) somewhere\n   (locally is fine too).\n\n\n1. Setup a `.env` file to configure Commodore (don\'t use quotes):\n\n   ```shell\n   # URL of Lieutenant API\n   COMMODORE_API_URL=https://lieutenant-api.example.com/\n   # Lieutenant API token\n   COMMODORE_API_TOKEN=<my-token>\n   # Your local user ID to be used in the container (optional, defaults to root)\n   USER_ID=<your-user-id>\n   # Your username to be used in the commits (optional, defaults to your local git config)\n   COMMODORE_USERNAME=<your name>\n   # Your user email to be used in the commits (optional, defaults to your local git config)\n   COMMODORE_USERMAIL=<your email>\n   ```\n1. Run commodore\n    ```console\n    commodore\n    ```\n\n## Run Commodore with poetry\n\n### Additional System Requirements\n\n* [Poetry](https://github.com/python-poetry/poetry) 1.1.0+\n* Docker\n\n\n1. Install requirements\n\n   Install poetry according to the upstream\n   [documentation](https://github.com/python-poetry/poetry#installation).\n\n   Create the Commodore environment:\n\n    ```console\n    poetry install\n    ```\n\n    Install jsonnet-bundler according to upstream [documentation](https://github.com/jsonnet-bundler/jsonnet-bundler#install).\n\n\n1. Finish setup as described [above](#getting_started_jsonnet)\n\n1. Run Commodore\n\n   ```console\n   poetry run commodore\n   ```\n\n1. Start hacking on Commodore\n\n   ```console\n   poetry shell\n   ```\n\n   - Write a line of test code, make the test fail\n   - Write a line of application code, make the test pass\n   - Repeat\n\n   Note: Commodore uses the [Black](https://github.com/psf/black) code formatter, and its formatting is encforced by CI.\n\n1. Run linting and tests\n\n   Auto format with autopep8\n   ```console\n   poetry run autopep\n   ```\n\n   List all Tox targets\n   ```console\n   poetry run tox -lv\n   ```\n\n   Run all linting and tests\n   ```console\n   poetry run tox\n   ```\n\n   Run just a specific target\n   ```console\n   poetry run tox -e py38\n   ```\n\n\n## Run Commodore in Docker\n\n**IMPORTANT:** After checking out this project, run `mkdir -p catalog inventory dependencies` in it before running any Docker commands.\nThis will ensure the folders are writable by the current user in the context of the Docker container.\n\nA docker-compose setup enables running Commodore in a container.\nThe environment variables are picked up from the local `.env` file.\nBy default your `~/.ssh/` directory is mounted into the container and an `ssh-agent` is started.\nYou can skip starting an agent by setting the `SSH_AUTH_SOCK` env variable and mounting the socket into the container.\n\n1. Build the Docker image inside of the cloned Commodore repository:\n\n```console\ndocker-compose build\n```\n\n1. Run the built image:\n\n```console\ndocker-compose run commodore catalog compile $CLUSTER_ID\n```\n\n## Documentation\n\nDocumentation for this component is written using [Asciidoc][asciidoc] and [Antora][antora].\nIt is located in the [docs/](docs) folder.\nThe [Divio documentation structure](https://documentation.divio.com/) is used to organize its content.\n\nRun the `make docs-serve` command in the root of the project, and then browse to http://localhost:2020 to see a preview of the current state of the documentation.\n\nAfter writing the documentation, please use the `make docs-vale` command and correct any warnings raised by the tool.\n\n## Contributing and license\n\nThis library is licensed under [BSD-3-Clause](LICENSE).\nFor information about how to contribute see [CONTRIBUTING](CONTRIBUTING.md).\n\n[asciidoc]: https://asciidoctor.org/\n[antora]: https://antora.org/\n',
    'author': 'VSHN AG',
    'author_email': 'info@vshn.ch',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/projectsyn/commodore',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



from __future__ import annotations

import shutil
import tempfile

from collections.abc import Iterable
from pathlib import Path as P
from textwrap import dedent
from typing import Optional

import click
import git
from kapitan.resources import inventory_reclass

from commodore.config import Config
from commodore.component import Component
from commodore.dependency_mgmt.component_library import (
    validate_component_library_name,
    create_component_library_aliases,
)
from commodore.dependency_mgmt.jsonnet_bundler import fetch_jsonnet_libraries
from commodore.helpers import kapitan_compile, relsymlink, yaml_dump
from commodore.inventory import Inventory
from commodore.inventory.lint import check_removed_reclass_variables
from commodore.multi_dependency import MultiDependency
from commodore.postprocess import postprocess_components


# pylint: disable=too-many-arguments disable=too-many-locals
def compile_component(
    config: Config,
    component_path_: str,
    instance_name_: Optional[str],
    value_files_: Iterable[str],
    search_paths_: Iterable[str],
    output_path_: str,
    repo_directory_: str,
    component_name: str,
):
    # Resolve all input to absolute paths to fix symlinks
    component_path = P(component_path_).resolve()
    value_files = [P(f).resolve() for f in value_files_]
    search_paths = [P(d).resolve() for d in search_paths_]
    search_paths.append(component_path / "vendor")
    output_path = P(output_path_).resolve()
    if repo_directory_:
        repo_directory = P(repo_directory_).resolve()
        sub_path = str(component_path.relative_to(repo_directory))
    else:
        # If no repo directory given, assume component path is repo root
        repo_directory = component_path
        sub_path = ""

    if not component_name:
        # Ignore 'component-' prefix in repo name, this assumes that the repo name
        # indicates the component name for components in subpaths.
        component_name = repo_directory.stem.replace("component-", "")

    # Fall back to `component as component` when instance_name is empty
    if instance_name_ is None or instance_name_ == "":
        instance_name = component_name
        click.secho(f"Compile component {component_name}...", bold=True)
    else:
        instance_name = instance_name_
        click.secho(
            f"Compile component {component_name} as {instance_name}...", bold=True
        )

    temp_dir = P(tempfile.mkdtemp(prefix="component-")).resolve()
    config.work_dir = temp_dir
    try:
        if config.debug:
            click.echo(f"   > Created temp workspace: {config.work_dir}")
        inv = config.inventory
        inv.ensure_dirs()
        search_paths.append(inv.dependencies_dir)
        component = _setup_component(
            config,
            component_name,
            instance_name,
            repo_directory,
            sub_path,
        )
        _prepare_kapitan_inventory(inv, component, value_files, instance_name)

        # Raise error if component uses removed reclass parameters
        check_removed_reclass_variables(
            config,
            "component",
            [component.defaults_file, component.class_file] + value_files,
        )

        # Verify component alias
        nodes = inventory_reclass(inv.inventory_dir)["nodes"]
        config.verify_component_aliases(nodes[instance_name]["parameters"])

        cluster_params = nodes[instance_name]["parameters"]
        create_component_library_aliases(config, cluster_params)

        # Render jsonnetfile.jsonnet if necessary
        component_params = nodes[instance_name]["parameters"].get(
            component_name.replace("-", "_"), {}
        )
        component.render_jsonnetfile_json(component_params)
        # Fetch Jsonnet libs
        fetch_jsonnet_libraries(component_path)

        # Compile component
        kapitan_compile(
            config,
            [instance_name],
            output_dir=output_path,
            search_paths=search_paths,
            fake_refs=True,
            reveal=True,
        )
        click.echo(
            f" > Component compiled to {output_path / 'compiled' / instance_name}"
        )

        # Change working directory for postprocessing
        config.work_dir = output_path
        postprocess_components(config, nodes, config.get_components())
        config.print_deprecation_notices()
    finally:
        if config.trace:
            click.echo(f" > Temp dir left in place {temp_dir}")
        else:
            if config.debug:
                click.echo(f" > Remove temp dir {temp_dir}")
            shutil.rmtree(temp_dir)


def _setup_component(
    config: Config,
    component_name: str,
    instance_name: str,
    repo_directory: P,
    sub_path: str,
) -> Component:
    if not repo_directory.is_dir():
        raise click.ClickException(
            f"Can't compile component, repository {repo_directory} doesn't exist"
        )
    cr = git.Repo(repo_directory)
    cdep = MultiDependency(cr.remote().url, config.inventory.dependencies_dir)
    component = Component(
        component_name, dependency=cdep, directory=repo_directory, sub_path=sub_path
    )
    config.register_component(component)
    config.register_component_aliases({instance_name: component_name})

    # Validate component libraries
    for lib in component.lib_files:
        validate_component_library_name(component.name, lib)

    return component


def _prepare_kapitan_inventory(
    inv: Inventory, component: Component, value_files: Iterable[P], instance_name: str
):
    """
    Setup Kapitan inventory.

    Create component symlinks, values file symlinks, setup params class with fake values
    and Kapitan target for the component, create a fake `lib/argocd.libjsonnet`.
    """
    component_class_file = component.class_file
    component_defaults_file = component.defaults_file
    if not component_class_file.exists():
        raise click.ClickException(
            f"Could not find component class file: {component_class_file}"
        )
    if not component_defaults_file.exists():
        raise click.ClickException(
            f"Could not find component default file: {component_defaults_file}"
        )

    # Create class symlink
    relsymlink(component_class_file, inv.components_dir)
    # Create defaults symlink
    relsymlink(
        component_defaults_file,
        inv.defaults_dir,
        dest_name=f"{component.name}.yml",
    )
    # Create component symlink
    relsymlink(component.target_directory, inv.dependencies_dir, component.name)
    # Create value symlinks
    for file in value_files:
        relsymlink(file.parent / file.name, inv.classes_dir)

    # Create class for fake parameters
    yaml_dump(
        {
            "parameters": {
                "cluster": {
                    "catalog_url": "ssh://git@git.example.com/org/repo.git",
                    "name": "c-green-test-1234",
                    "tenant": "t-silent-test-1234",
                },
                "facts": {
                    "distribution": "test-distribution",
                    "cloud": "cloudscale",
                    "region": "rma1",
                },
                "argocd": {
                    "namespace": "test",
                },
                "components": {
                    component.name: {
                        "url": "https://example.com/{component.name}.git",
                        "version": "master",
                    }
                },
                "kapitan": {
                    "vars": {
                        "target": instance_name,
                    },
                    "namespace": "test",
                },
            }
        },
        inv.params_file,
    )

    # Create test target
    value_classes = [f"{c.stem}" for c in value_files]
    yaml_dump(
        {
            "classes": [
                f"params.{inv.bootstrap_target}",
                f"defaults.{component.name}",
                f"components.{component.name}",
            ]
            + value_classes,
            "parameters": {
                "_instance": instance_name,
                "_base_directory": str(component.target_directory),
            },
        },
        inv.target_file(instance_name),
    )

    # Fake Argo CD lib
    # We plug "fake" Argo CD library here because every component relies on it
    # and we don't want to provide it every time when compiling a single component.
    with open(inv.lib_dir / "argocd.libjsonnet", "w", encoding="utf-8") as argocd_libf:
        argocd_libf.write(
            dedent(
                """
            local ArgoApp(component, namespace, project='', secrets=true) = {};
            local ArgoProject(name) = {};

            {
              App: ArgoApp,
              Project: ArgoProject,
            }"""
            )
        )



from __future__ import annotations

from collections.abc import Iterable
from typing import Callable


def format_component_list(
    components: Iterable[str], format_func: Callable[[str], str] = lambda c: f"'{c}'"
) -> str:
    formatted_list = list(map(format_func, sorted(components)))

    if len(formatted_list) == 0:
        return ""

    if len(formatted_list) == 1:
        return f"{formatted_list[0]}"

    formatted = ", ".join(formatted_list[:-1])

    # Use serial ("Oxford") comma when formatting lists of 3 or more items, cf.
    # https://en.wikipedia.org/wiki/Serial_comma
    serial_comma = ""
    if len(formatted_list) > 2:
        serial_comma = ","

    formatted += f"{serial_comma} and {formatted_list[-1]}"

    return formatted



from __future__ import annotations

from typing import Any
from pathlib import Path

import click

from commodore.config import Config
from commodore.helpers import relsymlink

from .tools import format_component_list


def validate_component_library_name(cname: str, lib: Path) -> Path:
    if not lib.stem.startswith(cname):
        raise click.ClickException(
            f"Component '{cname}' uses invalid component library name '{lib.name}'. "
            + "Consider using a library alias."
        )

    return lib


def _check_library_alias_prefixes(libalias: str, cn: str, component_prefixes: set[str]):
    for p in component_prefixes - {cn}:
        if libalias.startswith(p):
            raise click.ClickException(
                f"Invalid alias prefix '{p}' "
                + f"for template library alias of component '{cn}'"
            )


def _check_library_alias_collisions(cfg: Config, cluster_params: dict[str, Any]):
    # map of library alias to set(originating components)
    collisions: dict[str, set[str]] = {}

    components = cfg.get_components()
    component_prefixes = set(cluster_params["components"].keys())

    for cn, component in components.items():
        cmeta = cluster_params[component.parameters_key].get("_metadata", {})
        aliases = cmeta.get("library_aliases", {})
        for libalias in aliases.keys():
            _check_library_alias_prefixes(libalias, cn, component_prefixes)
            collisions.setdefault(libalias, set()).add(cn)

    for libalias, cnames in collisions.items():
        if len(cnames) > 1:
            clist = format_component_list(cnames)
            _all = "all" if len(cnames) > 2 else "both"
            raise click.ClickException(
                f"Components {clist} {_all} define component library alias '{libalias}'"
            )


def create_component_library_aliases(cfg: Config, cluster_params: dict[str, Any]):
    _check_library_alias_collisions(cfg, cluster_params)

    for _, component in cfg.get_components().items():
        cmeta = cluster_params[component.parameters_key].get("_metadata", {})
        aliases = cmeta.get("library_aliases", {}).items()

        for libalias, libname in aliases:
            if cfg.debug:
                click.echo(f"     > aliasing template library {libname} to {libalias}")
            libf = component.get_library(libname)
            if not libf:
                click.secho(
                    f" > [WARN] '{component.name}' template library alias '{libalias}' "
                    + f"refers to nonexistent template library '{libname}'",
                    fg="yellow",
                )
            else:
                relsymlink(libf, cfg.inventory.lib_dir, dest_name=libalias)
